/* ------------------------------------------------------------------------- */
/*
 *  sockstream.h
 *
 *  Copyright (c) 2004 - 2005 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Sat 05 Nov 2005 03:14:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_SOCKSTREAM_H
#define CLX_SOCKSTREAM_H

#include <iostream>
#include <streambuf>
#include <string>
#include <vector>
#include "clx/socket.h"
#include "clx/socktransmit_func.h"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  basic_sockbuf
	/* --------------------------------------------------------------------- */
	template <
		class TransmitFunc,
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_sockbuf : public std::basic_streambuf<CharT, Traits> {
	public:
		typedef CharT char_type;
		typedef TransmitFunc func;
		typedef Traits traits;
		typedef unsigned int size_type;
		typedef typename std::vector<char_type> container;
		typedef typename Traits::int_type int_type;
		
		// constructor and destructor
		explicit basic_sockbuf(socket_type s, const func& f = func(), size_type size) :
			std::basic_streambuf<CharT, Traits>(), s_(s), buffer_(size), f_(f) {}
		
		virtual ~basic_sockbuf() {}
		
		// operator
		bool is_open() const { return s_ >= 0; }
		
	protected:
		virtual int_type overflow(int_type c = traits::eof()) {
			if (!traits::not_eof(c)) return traits::eof();
			char tmp = static_cast<char>(c);
			if (this->xsputn(&tmp, 1) == 0) return traits::eof();
			return c;
		}
		
		virtual int_type underflow() {
			if (!this->is_open()) return traits::eof();
			if (this->gptr() >= this->egptr()) {
				if (this->gptr() > this->eback()) buffer_[0] = *(this->gptr() - 1);
				
				int l = this->xsgetn(&buffer_[1], static_cast<int>(buffer_.size() - 1));
				if (l <= 0) return traits::eof();
				
				this->setg(&buffer_[0], &buffer_[1], &buffer_[l + 1]);
			}
			return *this->gptr();
		}
		
		virtual std::streamsize xsputn(const char_type* s, std::streamsize n) {
			if (!this->is_open()) return 0;
			int l = f_.send(s_, s, n);
			return (l < 0) ? 0 : l;
		}
		
		virtual std::streamsize xsgetn(char_type* s, std::streamsize n) {
			if (!this->is_open()) return 0;
			int l = f_.recv(s_, s, n);
			return (l < 0) ? 0 : l;
		}
	
	private:
		socket_type s_;
		container buffer_;
		func f_;
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_sockstream
	/* --------------------------------------------------------------------- */
	template <
		class TransmitFunc,
		class CharT = char,
		class Traits = std::char_traits<CharT>
	>
	class basic_sockstream : public std::basic_iostream<CharT, Traits> {
	public:
		typedef unsigned int size_type;
		typedef CharT char_type;
		typedef TransmitFunc func;
		typedef basic_sockbuf<TransmitFunc, CharT, Traits> sockbuf;
		
		// constructor and destructor
		template <class Socket>
		explicit basic_sockstream(const Socket& s, const func& f = func(), size_type size = 4096) :
			std::basic_iostream<CharT, Traits>(0), sbuf_(s.socket(), f, size) {
			this->rdbuf(&sbuf_);
			if (!sbuf_.is_open()) this->setstate(std::ios::badbit);
		}
		
		explicit basic_sockstream(socket_type s, const func& f = func(), size_type size = 4096) :
			std::basic_iostream<CharT, Traits>(0), sbuf_(s, f, size) {
			this->rdbuf(&sbuf_);
			if (!sbuf_.is_open()) this->setstate(std::ios::badbit);
		}
		
		virtual ~basic_sockstream() { this->rdbuf(0); }
		
		// operator
		bool is_open() const { return this->rdbuf()->is_open(); }
		int send(const char_type* s, int n) { return this->rdbuf()->sputn(s, n); }
		int recv(char_type* s, int n) { return this->rdbuf()->sgetn(s, n); }
	
	private:
		sockbuf sbuf_;
	};
	
	typedef basic_sockstream<tcptransmit<char> > sockstream;
}

#endif // CLX_TCPSTREAM_H
