/* ------------------------------------------------------------------------- */
/*
 *  timefmt.h
 *
 *  Copyright (c) 2004 - 2006 "clown", All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Mon 27 Feb 2006 20:13:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_TIMEFMT_H
#define CLX_TIMEFMT_H

#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <map>
#include <string>
#include <sstream>
#include "lexical_cast.h"
#include "time.h"

#define TIMEFMT_STR "%y/%m/%d %H:%M:%S"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  basic_timefmt
	/* --------------------------------------------------------------------- */
	template <
		class CharT = char,
		class Traits = std::char_traits<CharT>
	>
	class basic_timefmt {
	public:
		typedef time time_type;
		typedef time::value_type value_type;
		typedef CharT char_type;
		
		typedef typename std::basic_string<CharT, Traits> string_type;
		typedef typename std::basic_stringstream<CharT, Traits> stream_type;
		typedef typename std::map<CharT, value_type*> time_map;
		typedef typename string_type::iterator iterator;
		
		static const char_type rep = '%';
		static const char_type ends = '\0';
		static const char_type ws = ' ';
		static const char_type lpos = '-';
		
		// constructor and destructor
		explicit basic_timefmt(const string_type& fmt = TIMEFMT_STR) :
			t_(), tmap_(), fmt_(fmt) { this->init(); }
		
		explicit basic_timefmt(const time_type& t, const string_type& fmt = TIMEFMT_STR) :
			t_(t), tmap_(), fmt_(fmt) { this->init(); }
		
		virtual ~basic_timefmt() {}
		
		// operator
		virtual void reset() { t_ = time_type(); }
		
		virtual time_type& assign(const string_type& s) {
			string_type tmp(s);
			
			iterator ps = tmp.begin();
			iterator pf = fmt_.begin();
			if (ps == tmp.end() || pf == fmt_.end()) return t_;
			
			do {
				if (*ps == ws) ps++; // skip white space
				else if (*pf == rep && *ps != *pf) {
					string_type flag;
					/*
					 * validate identifier in input method (assign()) is
					 * "%[0-9]*<id>" (<id> = y, m, d, ...). '-' and ' ' flags
					 * are only used by output method (str()).
					 */
					while (++pf != fmt_.end() && tmap_.find(*pf) == tmap_.end()) flag += *pf;
					if (pf == fmt_.end()) break; // no match
					char_type c = (pf + 1 != fmt_.end()) ? *(pf + 1) : ends;
					*tmap_[*pf++] = this->get(ps, tmp.end(), c, flag);
					if (ps++ == tmp.end()) break;
					pf++;
				}
				else if (*ps == *pf) { ps++; pf++; }
				else break; // no match
			} while (ps != tmp.end() && pf != fmt_.end());
			return t_;
		}
		
		virtual time_type& assign(const string_type& s, const string_type& fmt) {
			fmt_ = fmt;
			return this->assign(s);
		}
		
		virtual string_type str() {
			stream_type ss;
			for (iterator pf = fmt_.begin(); pf != fmt_.end(); pf++) {
				// if the last character is '%', treat this as same as "%%".
				if (*pf == rep && pf + 1 != fmt_.end() && *(pf + 1) != rep) {
					string_type flag;
					while (++pf != fmt_.end() && tmap_.find(*pf) == tmap_.end()) flag += *pf;
					if (pf == fmt_.end()) break; // uncorrect format
					this->put(ss, *tmap_[*pf], flag);
				}
				else ss << *pf;
			}
			return ss.str();
		}
		
		virtual string_type str(const string_type& fmt) {
			fmt_ = fmt;
			return this->str();
		}
		
		const time_type& data() const { return t_; }
		
	private:
		time_type t_;
		time_map tmap_;
		string_type fmt_;
		
		void init() {
			tmap_['y'] = &t_.year;
			tmap_['m'] = &t_.mon;
			tmap_['d'] = &t_.day;
			tmap_['H'] = &t_.hour;
			tmap_['M'] = &t_.min;
			tmap_['S'] = &t_.sec;
		}
		
		value_type get(iterator& next, iterator last, char_type c, const string_type& flag) {
			int n = -1;
			if (!flag.empty() && isdigit(flag.at(0))) n = lexical_cast<int>(flag);
			
			string_type tmp;
			while (next != last && *next != c && n-- != 0) {
				tmp += *next;
				next++;
			}
			if (tmp.empty()) return 0;
			return lexical_cast<value_type>(tmp);
		}
		
		void put(stream_type& ss, value_type val, const string_type& flag) {
			typedef typename string_type::size_type size_type;
			string_type tmp(flag);
			for (size_type i = 0; i < tmp.size(); i++) {
				if (tmp.at(i) == lpos) ss << std::left;
				else if (tmp.at(i) == ws || tmp.at(i) == '0') ss << std::setfill(tmp.at(i));
				else {
					ss << std::setw(lexical_cast<int>(tmp.substr(i)));
					break;
				}
			}
			ss << val;
		}
	};
	
	typedef basic_timefmt<> timefmt;
}

#endif // CLX_TIMEFMT_H
